import { resolve } from "path"
import invariant from "tiny-invariant"
import type { Stats } from "fs"

/**
 * An object that implements basic file system functionalities.
 *
 */
export interface IFileSystem {
  /**
   * Synchronously tests whether or not the given path exists by checking with the file system.
   * @param path A path to a file or directory.
   */
  existsSync(path: string): boolean

  /**
   * Asynchronously reads the entire contents of a file.
   * @param path A path to a file.
   */
  readFile(path: string, encoding: BufferEncoding): Promise<string>

  /**
   * Asynchronously writes data to a file, replacing the file if it already exists.
   * @param path A path to a file.
   * @param data The data to write.
   */
  writeFile(path: string, data: string): Promise<void>

  /**
   * Asynchronous readdir(3) - read a directory.
   * @param path A path to a directory.
   */
  readdir(path: string): Promise<string[]>

  /**
   * Asynchronous stat(2) - Get file status.
   * @param path A path to a file.
   */
  stat(path: string): Promise<Stats>
}

export function getDefaultFileSystem(): IFileSystem {
  const fs: typeof import("fs") = require("fs")
  return { ...fs.promises, existsSync: fs.existsSync }
}

export class OnePasswordFileManager {
  private root: string

  constructor(private fs: IFileSystem, path: string, profileName: string) {
    this.root = resolve(path, profileName)
    invariant(fs.existsSync(path), `Path ${path} does not exist.`)
    invariant(fs.existsSync(this.root), `Profile ${profileName} does not exist.`)
  }

  #hasFile(path: string) {
    return this.fs.existsSync(resolve(this.root, path))
  }

  async #readFile(path: string) {
    return await this.fs.readFile(resolve(this.root, path), "utf-8")
  }

  async #writeFile(path: string, value: string) {
    return await this.fs.writeFile(resolve(this.root, path), value)
  }

  getProfile() {
    return this.#readFile("profile.js")
  }

  getFolders() {
    return this.#readFile("folders.js")
  }

  async getBand(name: string) {
    const fileName = `band_${name.toUpperCase()}.js`
    if (this.#hasFile(fileName)) {
      return await this.#readFile(fileName)
    }
  }

  setProfile(profile: string) {
    this.#writeFile("profile.js", profile)
  }

  setFolders(folders: string) {
    this.#writeFile("folders.js", folders)
  }

  setBand(name: string, band: string) {
    this.#writeFile(`band_${name}.js`, band)
  }
}
