import { resolve } from "path"
import type { IFileSystem } from "./fs"
import { getDefaultFileSystem } from "./fs"
import { Vault } from "./vault"

export type { Vault } from "./vault"
export { Category, FieldType } from "./models"

interface Options {
  /**
   * Path to `.opvault` directory
   */
  path: string

  /**
   * Underlying `fs` module. You can replace it with a wrapper of
   * `memfs` or any object that implements `IFileSystem`.
   */
  fs?: IFileSystem
}

/**
 * OnePassword instance
 */
export class OnePassword {
  readonly #path: string
  readonly #fs: IFileSystem

  constructor({ path, fs = getDefaultFileSystem() }: Options) {
    this.#fs = fs
    this.#path = path
  }

  /**
   * @returns A list of names of profiles of the current vault.
   */
  async getProfileNames() {
    const [fs, path] = [this.#fs, this.#path]
    const children = await fs.readdir(path)
    const profiles: string[] = []
    await Promise.all(
      children.map(async child => {
        const fullPath = resolve(path, child)
        const stats = await fs.stat(fullPath)
        if (stats.isDirectory() && fs.existsSync(resolve(fullPath, "profile.js"))) {
          profiles.push(child)
        }
      })
    )
    return profiles
  }

  /**
   * @returns A OnePassword Vault instance.
   */
  async getProfile(profileName: string) {
    return await Vault.of(this.#path, profileName, this.#fs)
  }
}
