import * as crypto from "crypto"
import { HMACAssertionError, invariant } from "./errors"
import type { IFileSystem } from "./fs"
import { OnePasswordFileManager } from "./fs"
import { i18n } from "./i18n"

import type {
  Profile,
  Band,
  Overview,
  EncryptedItem,
  Item,
  AttachmentMetadata,
} from "./types"

type FoldersMap = { [uuid: string]: Band }

/**
 * Main OnePassword Vault class
 */
export class Vault {
  // Ciphers
  #master?: Cipher
  #overview?: Cipher

  // File system interface
  #files: OnePasswordFileManager

  // Encrypted contents
  #profile: Profile
  #folders: FoldersMap
  #bands = new Map<string, Band>()

  // Decrypted contents with plain-texts
  #overviews = new Map<string, Overview>()
  #items = new Map<string, EncryptedItem>()

  private constructor(
    files: OnePasswordFileManager,
    profile: Profile,
    folders: FoldersMap,
    bands: Map<string, Band>
  ) {
    this.#files = files
    this.#profile = profile
    this.#folders = folders
    this.#bands = bands
  }

  /**
   * Create a new OnePassword Vault instance and read all bands.
   */
  static async of(path: string, profileName = "default", fs: IFileSystem) {
    const files = new OnePasswordFileManager(fs, path, profileName)
    const profile = JSON.parse(
      stripText(await files.getProfile(), /^var profile\s*=/, ";")
    )
    const folders = JSON.parse(stripText(await files.getFolders(), "loadFolders(", ");"))
    const bands = new Map<string, Band>()

    for (let i = 0; i < 36; i++) {
      const letter = i.toString(36)
      const source = await files.getBand(letter)
      bands.set(letter, source ? JSON.parse(stripText(source, "ld(", ");")) : {})
    }

    return new Vault(files, profile, folders, bands)
  }

  readonly overviews = Object.freeze({
    get: (condition: string | ((overview: Overview) => boolean)) => {
      this.#assertUnlocked()

      if (typeof condition === "string") {
        const title = condition
        condition = overview => overview.title === title
      }
      for (const value of this.#overviews.values()) {
        if (condition(value)) {
          return value
        }
      }
    },

    values: () => {
      this.#assertUnlocked()
      return Array.from(this.#overviews.values())
    },
  })

  /**
   * Unlock this OnePassword vault.
   * @param masterPassword User provided master password. Only the derived
   * master and overview key will be stored within the class.
   */
  unlock(masterPassword: string) {
    const profile = this.#profile
    const derivedKey = crypto.pbkdf2Sync(
      /* password */ masterPassword,
      /* salt */ toBuffer(profile.salt),
      /* iterations */ profile.iterations,
      /* keylen */ 64,
      /* digest */ "sha512"
    )

    const cipher = splitPlainText(derivedKey)

    // Derive master key and overview keys
    try {
      this.#master = decryptKeys(profile.masterKey, cipher)
      this.#overview = decryptKeys(profile.overviewKey, cipher)
    } catch (e) {
      if (e instanceof HMACAssertionError) {
        throw new Error(i18n.error.invalidPassword)
      }
      throw e
    }

    this.#readOverviews()
    return this
  }

  decryptAttachment(buffer: Buffer) {
    return decryptAttachment(buffer, this.#master!)
  }

  /**
   * Remove derived keys stored within the class instance.
   */
  lock() {
    this.#master = null!
    this.#overview = null!
    this.#overviews.clear()
    return this
  }

  get isLocked() {
    return this.#master?.key == null
  }

  #assertUnlocked() {
    if (this.isLocked) {
      throw new Error(i18n.error.vaultIsLocked)
    }
  }

  getItem(uuid: string) {
    this.#assertUnlocked()
    const encrypted = uuid ? this.#bands.get(uuid[0])![uuid] : undefined
    return encrypted && decryptItem(encrypted, this.#master!)
  }

  #readOverviews() {
    this.#assertUnlocked()
    this.#bands.forEach(value => {
      for (const [uuid, item] of Object.entries(value)) {
        const overview = decryptOverview(item, this.#overview!)
        overview.uuid = uuid
        this.#overviews.set(uuid, overview)
      }
    })
  }
}

function decryptOverview(item: EncryptedItem, overviewCipher: Cipher) {
  try {
    const overview = decryptOPData(toBuffer(item.o), overviewCipher)
    return JSON.parse(overview.toString("utf8")) as Overview
  } catch (e) {
    console.error(i18n.error.cannotDecryptOverviewItem)
    throw e
  }
}

function decryptItem(item: EncryptedItem, master: Cipher): Item {
  const k = toBuffer(item.k)
  const data = k.slice(0, -32)
  assertHMac(data, master.hmac, k.slice(-32))
  const derivedKey = decryptData(master.key, data.slice(0, 16), data.slice(16))

  const detail = decryptOPData(
    /* cipherText */ toBuffer(item.d),
    /* cipher */ splitPlainText(derivedKey)
  )
  return JSON.parse(detail.toString("utf-8"))
}

function decryptAttachment(item: Buffer, master: Cipher) {
  invariant(
    item.slice(0, 6).toString("utf-8") === "OPCLDA",
    "Attachment must start with OPCLDA"
  )
  invariant(
    item.readIntLE(7, 1) === 1,
    "The version for this attachment file format is not supported."
  )

  const metadataSize = item.readIntLE(8, 2)
  const iconSize = item.readIntLE(12, 3)

  const metadata: AttachmentMetadata = JSON.parse(
    item.slice(16, 16 + metadataSize).toString("utf-8")
  )
  const icondata = item.slice(16 + metadataSize, 16 + metadataSize + iconSize)
  console.log(icondata.slice(0, 8).toString())
  const iconData = decryptOPData(icondata, master)
  return iconData
}

/** Encryption and MAC */
interface Cipher {
  /** Encryption key */
  key: Buffer
  /** HMAC key */
  hmac: Buffer
}

function stripText(text: string, prefix: string | RegExp, suffix: string | RegExp) {
  if (typeof prefix === "string") {
    if (text.startsWith(prefix)) {
      text = text.slice(prefix.length)
    }
  } else {
    const prefixMatch = text.match(prefix)
    if (prefixMatch) {
      text = text.slice(prefixMatch[0].length)
    }
  }
  if (typeof suffix === "string") {
    if (text.endsWith(suffix)) {
      text = text.slice(0, -suffix.length)
    }
  } else {
    const suffixMatch = text.match(suffix)
    if (suffixMatch) {
      text = text.slice(0, -suffixMatch[0].length)
    }
  }
  return text
}

const splitPlainText = (derivedKey: Buffer): Cipher => ({
  key: derivedKey.slice(0, 32),
  hmac: derivedKey.slice(32, 64),
})

function decryptOPData(cipherText: Buffer, cipher: Cipher) {
  const key = cipherText.slice(0, -32)
  assertHMac(key, cipher.hmac, cipherText.slice(-32))

  const plaintext = decryptData(cipher.key, key.slice(16, 32), key.slice(32))
  const size = readUint16(key.slice(8, 16))
  return plaintext.slice(-size)
}

function assertHMac(data: Buffer, key: Buffer, expected: Buffer) {
  const actual = crypto.createHmac("sha256", key).update(data).digest()
  if (!actual.equals(expected)) {
    throw new HMACAssertionError()
  }
}

function decryptKeys(encryptedKey: string, derived: Cipher) {
  const buffer = toBuffer(encryptedKey)
  const base = decryptOPData(buffer, derived)
  const digest = crypto.createHash("sha512").update(base).digest()
  return splitPlainText(digest)
}

function decryptData(key: Buffer, iv: Buffer, data: Buffer) {
  const cipher = crypto.createDecipheriv("aes-256-cbc", key, iv).setAutoPadding(false)
  return Buffer.concat([cipher.update(data), cipher.final()])
}

function readUint16({ buffer, byteOffset, length }: Buffer) {
  return new DataView(buffer, byteOffset, length).getUint16(0, true)
}

function toBuffer(data: string) {
  return Buffer.from(data, "base64")
}
